/*
 * Copyright (C) 2004, 2005, 2006 Moritz Orbach <zufall@apfelboymchen.homeunix.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>

#include <X11/Xutil.h>
#include <Imlib2.h>

#include "config.h"

#include "list.h"
#include "changeimage.h"
#include "cfg.h"
#include "status.h"
#include "recurse.h"
#include "memman.h"

#ifdef FENCE
#include "efence.h"
#endif
#ifdef DEBUG
#include "memwatch.h"
#endif

/* vielleicht
 * - vernnftige hintergrundfarbe:
 *   . zufall ohne grelle farben
 *   . r, g und b max- bzw. minimalwerte festlegen
 * - datei auslesen in denen ausnahmen zu default optionen stehen (pro bild). ugh.
 */

t_list *imglist;
Display *Xdisplay = NULL;
Screen *scr;
Window Xroot;
int screen;

unsigned int getrandom(unsigned int max)
{
	if (max > 0)
		return(rand() % max + 1);
	else
		return 0;
} 

void exit_func(void)
{
	if (!list_destroy(imglist))
		fprintf(stderr, "error freeing memory..\n");

	if (Xdisplay)
		XCloseDisplay(Xdisplay);

#ifdef DEBUG
	memwatchstop();
	printf("rest: %d\n", allocated());
	if (imglist == NULL)
		printf("list == NULL\n");
	else
		printf("listcount: %d\n", list_count(imglist));
#endif

}

/*
 * stolen from xcompmgr
 */
static void give_me_a_name(void)
{
	Window w;

	w = XCreateSimpleWindow(Xdisplay, Xroot , 0, 0, 1, 1, 0, None, None);
	Xutf8SetWMProperties(Xdisplay, w, PACKAGE_NAME, PACKAGE_NAME, NULL, 0, NULL, NULL, NULL);
}

void initX(void)
{
	/* clients without windows don't appear in xlsclients */
	Xdisplay = XOpenDisplay(NULL);
	if (!Xdisplay) {
		fprintf(stderr, "%s: Unable to open display\n", PACKAGE_NAME);
		exit(EXIT_FAILURE);
	}

	if (atexit(exit_func)) {
		fprintf(stderr,"%s: atexit() failed\n", PACKAGE_NAME);
		exit(EXIT_FAILURE);
	}

	screen = DefaultScreen(Xdisplay);
	Xroot = RootWindow(Xdisplay, screen);
	scr = ScreenOfDisplay(Xdisplay, screen);
	give_me_a_name();
	imlib_context_set_display(Xdisplay);
	imlib_context_set_visual(DefaultVisual(Xdisplay, DefaultScreen(Xdisplay)));
	imlib_context_set_anti_alias(1);
	imlib_context_set_dither(1);
	imlib_context_set_blend(1);
}


/* signal handlers */

void infotrap(p)
{
	unsigned int tmp;

	tmp = list_allocated();

	/*
	 * "calling library functions from a signal handler is not guaranteed to work 
	 * in all implementations....."
	 * http://publications.gbdirect.co.uk/c_book/chapter9/signal_handling.html
	 * http://www.delorie.com/gnu/docs/glibc/libc_488.html
	 * hm..
	 */
	/* warning: ISO C90 does not support the `%lf' printf format */
	printf("allocated: %d b (%.2lf kb), %d nodes\n", tmp, (double)tmp / 1024, list_count(imglist));
	signal(SIGUSR1, infotrap);
	break_sleep(0);
}

void nop(p)
{
	/* NOP */
	break_sleep(1);
}

void signal_quit()
{
	if (running()) {
		printf("%s: exiting\n", PACKAGE_NAME);
		halt();
		break_sleep(1);
		if (Xdisplay) {
			set_permanent(Xdisplay, Xroot);
		}
	} else
		printf("%s: i heard you\n", PACKAGE_NAME);
}

void signal_initquit(p)
{
	signal(p, SIG_DFL);
	signal_quit();
	/* this also kills a blocking read(0, ...) */
	raise(p);
}


/* --------- */

int main(int argc, char *argv[])
{
	char *image;
	unsigned int zufall = 0;
	unsigned int tr;	/* time remaining */
	int error = 0;		/* return code */
	int diroffset = 0;
	char ***collections = NULL;
	int collectionsc = 0;
	int passes;
	int retain = 0;

	/* killall / CTRL-C */
	signal(SIGTERM, signal_initquit);
	signal(SIGINT, signal_initquit);

	/*
	 * don't buffer output more than a line, so that the filename can be 
	 * seen immediately with tail -f when the image changes.
	 * "The standard error stream stderr is always unbuffered by default."
	 */
	setvbuf(stdout, (char *)NULL, _IOLBF, 0);

	/* cfg_file(); */
	cfg_cmdln(&argc, argv, &diroffset, &collections, &collectionsc);

	argc = argc - diroffset;
	argv = &argv[diroffset];
	/* ugh */
	passes = cfg.passes;

	initX();
	imglist = list_init();
	srand(time (0));

	buildlist(argc, argv, collections, collectionsc);
	/* killall / CTRL-C */
	signal(SIGTERM, signal_quit);
	signal(SIGINT, signal_quit);

	/* free memory for pointers to collections */
	freemem(collections, sizeof(*collections));

	if (list_count(imglist) > 0) {
		/* imagechange, info */
		signal(SIGHUP, nop);
		signal(SIGUSR1, infotrap);
	} else {
		error = 1;
		halt();
		fprintf(stderr, "Sorry, no supported images found\n");
	}

	while (running() && !error) {

		if (!cfg.sequential)
			zufall = getrandom(list_count(imglist));
		else {
			zufall++;
			if (zufall > list_count(imglist))
					/* wrap around */
					zufall = 1;
		}
		image = list_get(imglist, zufall);
		if (image == NULL) {
			/* more than 0 images, but there is no filename?? */
			halt();
			fprintf(stderr, "%s: internal error. pardon!\n", PACKAGE_NAME);
			break;
		}

		printf("so ein zufall!: %s\n", image);

		/* don't sleep if there is only one image left */
		if (list_count(imglist) > 1 && passes != 1)
			tr = cfg.delay;
		else {
			tr = 0;
			printf("no pictures left, exiting\n");
			retain = 1;
			halt();
		}
		if (changeimage(Xdisplay, Xroot, scr, screen, image, retain)) {
			while (tr != 0) {
				/* admire background */
				tr = sleep(tr);
				/* sleep is over / signal. continue sleep or new sleep? */
				if (sleep_broken()) {
					/* imagechange or exit */
					tr = 0;
					break_sleep(0);
				} else {
					/* must be USR1 */
					if (tr != 0) {
						printf("%s\n", image);
						printf("%d seconds (%.1d:%.2d) remaining\n", tr, (int)tr / 60, (int)tr % 60);
					}
				}
			}
			if (cfg.passes != -1)
				passes--;
		} else {
			/* changeimage failed */
			fprintf(stderr, "Unable to load image file \"%s\".\n", image);
			if (!list_delete(imglist, zufall)) {
				fprintf(stderr, "error freeing memory..\n");
				error = 1;
			}
		}
	}

	/* make return-code system-independent */
	switch (error) {
		case  0: error = EXIT_SUCCESS;
			 break;
		default: error = EXIT_FAILURE;
	}
	exit(error);
}
